package jdk.internal.org.objectweb.asm.util;

import jdk.internal.org.objectweb.asm.Type;
import jdk.internal.org.objectweb.asm.Opcodes;
import java.util.Iterator;
import jdk.internal.org.objectweb.asm.Handle;
import jdk.internal.org.objectweb.asm.Attribute;
import jdk.internal.org.objectweb.asm.TypePath;
import jdk.internal.org.objectweb.asm.AnnotationVisitor;
import java.io.PrintWriter;
import java.io.StringWriter;
import jdk.internal.org.objectweb.asm.tree.analysis.Analyzer;
import jdk.internal.org.objectweb.asm.tree.analysis.BasicVerifier;
import jdk.internal.org.objectweb.asm.tree.MethodNode;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.lang.reflect.Field;
import java.util.List;
import java.util.Set;
import jdk.internal.org.objectweb.asm.Label;
import java.util.Map;
import jdk.internal.org.objectweb.asm.MethodVisitor;

public class CheckMethodAdapter extends MethodVisitor
{
    public int version;
    private int access;
    private boolean startCode;
    private boolean endCode;
    private boolean endMethod;
    private int insnCount;
    private final Map<Label, Integer> labels;
    private Set<Label> usedLabels;
    private int expandedFrames;
    private int compressedFrames;
    private int lastFrame;
    private List<Label> handlers;
    private static final int[] TYPE;
    private static Field labelStatusField;
    
    public CheckMethodAdapter(MethodVisitor methodVisitor) {
        this(methodVisitor, new HashMap());
    }
    
    public CheckMethodAdapter(MethodVisitor methodVisitor, Map<Label, Integer> map) {
        this(327680, methodVisitor, map);
        if (this.getClass() != CheckMethodAdapter.class) {
            throw new IllegalStateException();
        }
        return;
    }
    
    protected CheckMethodAdapter(int api, MethodVisitor methodVisitor, Map<Label, Integer> labels) {
        super(api, methodVisitor);
        this.lastFrame = -1;
        this.labels = labels;
        this.usedLabels = new HashSet();
        this.handlers = new ArrayList();
    }
    
    public CheckMethodAdapter(final int access, final String s, final String s2, final MethodVisitor methodVisitor, Map<Label, Integer> map) {
        this(new MethodNode(327680, access, s, s2, null, null) {
            @Override
            public void visitEnd() {
                Analyzer<BasicValue> analyzer = new Analyzer(new BasicVerifier());
                try {
                    analyzer.analyze("dummy", this);
                }
                catch (Exception ex) {
                    if (ex instanceof IndexOutOfBoundsException && this.maxLocals == 0 && this.maxStack == 0) {
                        throw new RuntimeException("Data flow checking option requires valid, non zero maxLocals and maxStack values.");
                    }
                    ex.printStackTrace();
                    StringWriter out = new StringWriter();
                    PrintWriter printWriter = new PrintWriter(out, true);
                    CheckClassAdapter.printAnalyzerResult(this, analyzer, printWriter);
                    printWriter.close();
                    throw new RuntimeException(ex.getMessage() + 32 + out.toString());
                }
                this.accept(methodVisitor);
            }
        }, map);
        this.access = access;
    }
    
    @Override
    public void visitParameter(String name, int access) {
        if (name != null) {
            checkUnqualifiedName(this.version, name, "name");
        }
        CheckClassAdapter.checkAccess(access, 36880);
        super.visitParameter(name, access);
    }
    
    @Override
    public AnnotationVisitor visitAnnotation(String descriptor, boolean visible) {
        this.checkEndMethod();
        checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitAnnotation(descriptor, visible));
    }
    
    @Override
    public AnnotationVisitor visitTypeAnnotation(int typeRef, TypePath typePath, String descriptor, boolean visible) {
        this.checkEndMethod();
        int i = typeRef >>> 24;
        if (i != 1 && i != 18 && i != 20 && i != 21 && i != 22 && i != 23) {
            throw new IllegalArgumentException("Invalid type reference sort 0x" + Integer.toHexString(i));
        }
        CheckClassAdapter.checkTypeRefAndPath(typeRef, typePath);
        checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitTypeAnnotation(typeRef, typePath, descriptor, visible));
    }
    
    @Override
    public AnnotationVisitor visitAnnotationDefault() {
        this.checkEndMethod();
        return new CheckAnnotationAdapter(super.visitAnnotationDefault(), false);
    }
    
    @Override
    public AnnotationVisitor visitParameterAnnotation(int parameter, String descriptor, boolean visible) {
        this.checkEndMethod();
        checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitParameterAnnotation(parameter, descriptor, visible));
    }
    
    @Override
    public void visitAttribute(Attribute attribute) {
        this.checkEndMethod();
        if (attribute == null) {
            throw new IllegalArgumentException("Invalid attribute (must not be null)");
        }
        super.visitAttribute(attribute);
        return;
    }
    
    @Override
    public void visitCode() {
        if ((this.access & 0x400) != 0x0) {
            throw new RuntimeException("Abstract methods cannot have code");
        }
        this.startCode = true;
        super.visitCode();
        return;
    }
    
    @Override
    public void visitFrame(int n, int n2, Object[] local, int n3, Object[] stack) {
        if (this.insnCount == this.lastFrame) {
            throw new IllegalStateException("At most one frame can be visited at a given code location.");
        }
        this.lastFrame = this.insnCount;
        int n4;
        int n5;
        switch (n) {
            case -1:
            case 0: {
                n4 = Integer.MAX_VALUE;
                n5 = Integer.MAX_VALUE;
                break;
            }
            case 3: {
                n4 = 0;
                n5 = 0;
                break;
            }
            case 4: {
                n4 = 0;
                n5 = 1;
                break;
            }
            case 1:
            case 2: {
                n4 = 3;
                n5 = 0;
                break;
            }
            default: {
                throw new IllegalArgumentException("Invalid frame type " + n);
            }
        }
        if (n2 > n4) {
            throw new IllegalArgumentException("Invalid nLocal=" + n2 + " for frame type " + n);
        }
        if (n3 > n5) {
            throw new IllegalArgumentException("Invalid nStack=" + n3 + " for frame type " + n);
        }
        if (n != 2) {
            if (n2 > 0 && (local == null || local.length < n2)) {
                throw new IllegalArgumentException("Array local[] is shorter than nLocal");
            }
            for (int i = 0; i < n2; ++i) {
                this.checkFrameValue(local[i]);
            }
        }
        if (n3 > 0 && (stack == null || stack.length < n3)) {
            throw new IllegalArgumentException("Array stack[] is shorter than nStack");
        }
        for (int j = 0; j < n3; ++j) {
            this.checkFrameValue(stack[j]);
        }
        if (n == -1) {
            ++this.expandedFrames;
        }
        else {
            ++this.compressedFrames;
        }
        if (this.expandedFrames > 0 && this.compressedFrames > 0) {
            throw new RuntimeException("Expanded and compressed frames must not be mixed.");
        }
        super.visitFrame(n, n2, local, n3, stack);
        return;
    }
    
    @Override
    public void visitInsn(int opcode) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 0);
        super.visitInsn(opcode);
        ++this.insnCount;
    }
    
    @Override
    public void visitIntInsn(int opcode, int n) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 1);
        switch (opcode) {
            case 16: {
                checkSignedByte(n, "Invalid operand");
                break;
            }
            case 17: {
                checkSignedShort(n, "Invalid operand");
                break;
            }
            default: {
                if (n < 4 || n > 11) {
                    throw new IllegalArgumentException("Invalid operand (must be an array type code T_...): " + n);
                }
                break;
            }
        }
        super.visitIntInsn(opcode, n);
        ++this.insnCount;
    }
    
    @Override
    public void visitVarInsn(int opcode, int varIndex) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 2);
        checkUnsignedShort(varIndex, "Invalid variable index");
        super.visitVarInsn(opcode, varIndex);
        ++this.insnCount;
    }
    
    @Override
    public void visitTypeInsn(int opcode, String s) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 3);
        checkInternalName(s, "type");
        if (opcode == 187 && s.charAt(0) == '[') {
            throw new IllegalArgumentException("NEW cannot be used to create arrays: " + s);
        }
        super.visitTypeInsn(opcode, s);
        ++this.insnCount;
        return;
    }
    
    @Override
    public void visitFieldInsn(int opcode, String owner, String name, String descriptor) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 4);
        checkInternalName(owner, "owner");
        checkUnqualifiedName(this.version, name, "name");
        checkDesc(descriptor, false);
        super.visitFieldInsn(opcode, owner, name, descriptor);
        ++this.insnCount;
    }
    
    @Deprecated
    @Override
    public void visitMethodInsn(int opcode, String owner, String name, String descriptor) {
        if (this.api >= 327680) {
            super.visitMethodInsn(opcode, owner, name, descriptor);
            return;
        }
        this.doVisitMethodInsn(opcode, owner, name, descriptor, opcode == 185);
        return;
    }
    
    @Override
    public void visitMethodInsn(int opcode, String owner, String name, String descriptor, boolean isInterface) {
        if (this.api < 327680) {
            super.visitMethodInsn(opcode, owner, name, descriptor, isInterface);
            return;
        }
        this.doVisitMethodInsn(opcode, owner, name, descriptor, isInterface);
        return;
    }
    
    private void doVisitMethodInsn(int opcode, String owner, String s, String descriptor, boolean isInterface) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 5);
        if (opcode != 183 || !"<init>".equals(s)) {
            checkMethodIdentifier(this.version, s, "name");
        }
        checkInternalName(owner, "owner");
        checkMethodDesc(descriptor);
        if (opcode == 182 && isInterface) {
            throw new IllegalArgumentException("INVOKEVIRTUAL can't be used with interfaces");
        }
        if (opcode == 185 && !isInterface) {
            throw new IllegalArgumentException("INVOKEINTERFACE can't be used with classes");
        }
        if (this.mv != null) {
            this.mv.visitMethodInsn(opcode, owner, s, descriptor, isInterface);
        }
        ++this.insnCount;
        return;
    }
    
    @Override
    public void visitInvokeDynamicInsn(String name, String descriptor, Handle bootstrapMethodHandle, Object... bootstrapMethodArguments) {
        this.checkStartCode();
        this.checkEndCode();
        checkMethodIdentifier(this.version, name, "name");
        checkMethodDesc(descriptor);
        if (bootstrapMethodHandle.getTag() != 6 && bootstrapMethodHandle.getTag() != 8) {
            throw new IllegalArgumentException("invalid handle tag " + bootstrapMethodHandle.getTag());
        }
        for (int i = 0; i < bootstrapMethodArguments.length; ++i) {
            this.checkLDCConstant(bootstrapMethodArguments[i]);
        }
        super.visitInvokeDynamicInsn(name, descriptor, bootstrapMethodHandle, bootstrapMethodArguments);
        ++this.insnCount;
        return;
    }
    
    @Override
    public void visitJumpInsn(int opcode, Label label) {
        this.checkStartCode();
        this.checkEndCode();
        checkOpcode(opcode, 6);
        this.checkLabel(label, false, "label");
        checkNonDebugLabel(label);
        super.visitJumpInsn(opcode, label);
        this.usedLabels.add(label);
        ++this.insnCount;
    }
    
    @Override
    public void visitLabel(Label label) {
        this.checkStartCode();
        this.checkEndCode();
        this.checkLabel(label, false, "label");
        if (this.labels.get(label) != null) {
            throw new IllegalArgumentException("Already visited label");
        }
        this.labels.put(label, Integer.valueOf(this.insnCount));
        super.visitLabel(label);
        return;
    }
    
    @Override
    public void visitLdcInsn(Object value) {
        this.checkStartCode();
        this.checkEndCode();
        this.checkLDCConstant(value);
        super.visitLdcInsn(value);
        ++this.insnCount;
    }
    
    @Override
    public void visitIincInsn(int varIndex, int increment) {
        this.checkStartCode();
        this.checkEndCode();
        checkUnsignedShort(varIndex, "Invalid variable index");
        checkSignedShort(increment, "Invalid increment");
        super.visitIincInsn(varIndex, increment);
        ++this.insnCount;
    }
    
    @Override
    public void visitTableSwitchInsn(int n, int n2, Label dflt, Label... labels) {
        this.checkStartCode();
        this.checkEndCode();
        if (n2 < n) {
            throw new IllegalArgumentException("Max = " + n2 + " must be greater than or equal to min = " + n);
        }
        this.checkLabel(dflt, false, "default label");
        checkNonDebugLabel(dflt);
        if (labels == null || labels.length != n2 - n + 1) {
            throw new IllegalArgumentException("There must be max - min + 1 labels");
        }
        for (int i = 0; i < labels.length; ++i) {
            this.checkLabel(labels[i], false, "label at index " + i);
            checkNonDebugLabel(labels[i]);
        }
        super.visitTableSwitchInsn(n, n2, dflt, labels);
        for (int j = 0; j < labels.length; ++j) {
            this.usedLabels.add(labels[j]);
        }
        ++this.insnCount;
        return;
    }
    
    @Override
    public void visitLookupSwitchInsn(Label dflt, int[] keys, Label[] labels) {
        this.checkEndCode();
        this.checkStartCode();
        this.checkLabel(dflt, false, "default label");
        checkNonDebugLabel(dflt);
        if (keys == null || labels == null || keys.length != labels.length) {
            throw new IllegalArgumentException("There must be the same number of keys and labels");
        }
        for (int i = 0; i < labels.length; ++i) {
            this.checkLabel(labels[i], false, "label at index " + i);
            checkNonDebugLabel(labels[i]);
        }
        super.visitLookupSwitchInsn(dflt, keys, labels);
        this.usedLabels.add(dflt);
        for (int j = 0; j < labels.length; ++j) {
            this.usedLabels.add(labels[j]);
        }
        ++this.insnCount;
        return;
    }
    
    @Override
    public void visitMultiANewArrayInsn(String s, int i) {
        this.checkStartCode();
        this.checkEndCode();
        checkDesc(s, false);
        if (s.charAt(0) != '[') {
            throw new IllegalArgumentException("Invalid descriptor (must be an array type descriptor): " + s);
        }
        if (i < 1) {
            throw new IllegalArgumentException("Invalid dimensions (must be greater than 0): " + i);
        }
        if (i > s.lastIndexOf(91) + 1) {
            throw new IllegalArgumentException("Invalid dimensions (must not be greater than dims(desc)): " + i);
        }
        super.visitMultiANewArrayInsn(s, i);
        ++this.insnCount;
        return;
    }
    
    @Override
    public AnnotationVisitor visitInsnAnnotation(int typeRef, TypePath typePath, String descriptor, boolean visible) {
        this.checkStartCode();
        this.checkEndCode();
        int i = typeRef >>> 24;
        if (i != 67 && i != 68 && i != 69 && i != 70 && i != 71 && i != 72 && i != 73 && i != 74 && i != 75) {
            throw new IllegalArgumentException("Invalid type reference sort 0x" + Integer.toHexString(i));
        }
        CheckClassAdapter.checkTypeRefAndPath(typeRef, typePath);
        checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitInsnAnnotation(typeRef, typePath, descriptor, visible));
    }
    
    @Override
    public void visitTryCatchBlock(Label start, Label end, Label handler, String type) {
        this.checkStartCode();
        this.checkEndCode();
        this.checkLabel(start, false, "start label");
        this.checkLabel(end, false, "end label");
        this.checkLabel(handler, false, "handler label");
        checkNonDebugLabel(start);
        checkNonDebugLabel(end);
        checkNonDebugLabel(handler);
        if (this.labels.get(start) != null || this.labels.get(end) != null || this.labels.get(handler) != null) {
            throw new IllegalStateException("Try catch blocks must be visited before their labels");
        }
        if (type != null) {
            checkInternalName(type, "type");
        }
        super.visitTryCatchBlock(start, end, handler, type);
        this.handlers.add(start);
        this.handlers.add(end);
        return;
    }
    
    @Override
    public AnnotationVisitor visitTryCatchAnnotation(int typeRef, TypePath typePath, String descriptor, boolean visible) {
        this.checkStartCode();
        this.checkEndCode();
        int i = typeRef >>> 24;
        if (i != 66) {
            throw new IllegalArgumentException("Invalid type reference sort 0x" + Integer.toHexString(i));
        }
        CheckClassAdapter.checkTypeRefAndPath(typeRef, typePath);
        checkDesc(descriptor, false);
        return new CheckAnnotationAdapter(super.visitTryCatchAnnotation(typeRef, typePath, descriptor, visible));
    }
    
    @Override
    public void visitLocalVariable(String name, String descriptor, String signature, Label start, Label end, int index) {
        this.checkStartCode();
        this.checkEndCode();
        checkUnqualifiedName(this.version, name, "name");
        checkDesc(descriptor, false);
        this.checkLabel(start, true, "start label");
        this.checkLabel(end, true, "end label");
        checkUnsignedShort(index, "Invalid variable index");
        if ((int)this.labels.get(end) < (int)this.labels.get(start)) {
            throw new IllegalArgumentException("Invalid start and end labels (end must be greater than start)");
        }
        super.visitLocalVariable(name, descriptor, signature, start, end, index);
        return;
    }
    
    @Override
    public AnnotationVisitor visitLocalVariableAnnotation(int typeRef, TypePath typePath, Label[] start, Label[] end, int[] index, String descriptor, boolean visible) {
        this.checkStartCode();
        this.checkEndCode();
        int i = typeRef >>> 24;
        if (i != 64 && i != 65) {
            throw new IllegalArgumentException("Invalid type reference sort 0x" + Integer.toHexString(i));
        }
        CheckClassAdapter.checkTypeRefAndPath(typeRef, typePath);
        checkDesc(descriptor, false);
        if (start == null || end == null || index == null || end.length != start.length || index.length != start.length) {
            throw new IllegalArgumentException("Invalid start, end and index arrays (must be non null and of identical length");
        }
        int j = 0;
        while (j < start.length) {
            this.checkLabel(start[j], true, "start label");
            this.checkLabel(end[j], true, "end label");
            checkUnsignedShort(index[j], "Invalid variable index");
            if ((int)this.labels.get(end[j]) < (int)this.labels.get(start[j])) {
                throw new IllegalArgumentException("Invalid start and end labels (end must be greater than start)");
            }
            ++j;
        }
        return super.visitLocalVariableAnnotation(typeRef, typePath, start, end, index, descriptor, visible);
    }
    
    @Override
    public void visitLineNumber(int line, Label start) {
        this.checkStartCode();
        this.checkEndCode();
        checkUnsignedShort(line, "Invalid line number");
        this.checkLabel(start, true, "start label");
        super.visitLineNumber(line, start);
    }
    
    @Override
    public void visitMaxs(int maxStack, int maxLocals) {
        this.checkStartCode();
        this.checkEndCode();
        this.endCode = true;
        Iterator<Label> iterator = this.usedLabels.iterator();
        while (iterator.hasNext()) {
            if (this.labels.get((Object)iterator.next()) == null) {
                throw new IllegalStateException("Undefined label used");
            }
            continue;
        }
        int i = 0;
        while (i < this.handlers.size()) {
            Integer n = (Integer)this.labels.get(this.handlers.get(i++));
            Integer n2 = (Integer)this.labels.get(this.handlers.get(i++));
            if (n == null || n2 == null) {
                throw new IllegalStateException("Undefined try catch block labels");
            }
            if ((int)n2 <= (int)n) {
                throw new IllegalStateException("Emty try catch block handler range");
            }
            continue;
        }
        checkUnsignedShort(maxStack, "Invalid max stack");
        checkUnsignedShort(maxLocals, "Invalid max locals");
        super.visitMaxs(maxStack, maxLocals);
    }
    
    @Override
    public void visitEnd() {
        this.checkEndMethod();
        this.endMethod = true;
        super.visitEnd();
    }
    
    void checkStartCode() {
        if (!this.startCode) {
            throw new IllegalStateException("Cannot visit instructions before visitCode has been called.");
        }
        return;
    }
    
    void checkEndCode() {
        if (this.endCode) {
            throw new IllegalStateException("Cannot visit instructions after visitMaxs has been called.");
        }
        return;
    }
    
    void checkEndMethod() {
        if (this.endMethod) {
            throw new IllegalStateException("Cannot visit elements after visitEnd has been called.");
        }
        return;
    }
    
    void checkFrameValue(Object obj) {
        if (obj == Opcodes.TOP || obj == Opcodes.INTEGER || obj == Opcodes.FLOAT || obj == Opcodes.LONG || obj == Opcodes.DOUBLE || obj == Opcodes.NULL || obj == Opcodes.UNINITIALIZED_THIS) {
            return;
        }
        if ((obj instanceof String)) {
            checkInternalName((String)obj, "Invalid stack frame value");
            return;
        }
        if (!(obj instanceof Label)) {
            throw new IllegalArgumentException("Invalid stack frame value: " + obj);
        }
        this.usedLabels.add((Label)obj);
        return;
    }
    
    static void checkOpcode(int i, int n) {
        if (i < 0 || i > 199 || CheckMethodAdapter.TYPE[i] != n) {
            throw new IllegalArgumentException("Invalid opcode: " + i);
        }
        return;
    }
    
    static void checkSignedByte(int i, String str) {
        if (i < -128 || i > 127) {
            throw new IllegalArgumentException(str + " (must be a signed byte): " + i);
        }
        return;
    }
    
    static void checkSignedShort(int i, String str) {
        if (i < -32768 || i > 32767) {
            throw new IllegalArgumentException(str + " (must be a signed short): " + i);
        }
        return;
    }
    
    static void checkUnsignedShort(int i, String str) {
        if (i < 0 || i > 65535) {
            throw new IllegalArgumentException(str + " (must be an unsigned short): " + i);
        }
        return;
    }
    
    static void checkConstant(Object obj) {
        if (!(obj instanceof Integer) && !(obj instanceof Float) && !(obj instanceof Long) && !(obj instanceof Double) && !(obj instanceof String)) {
            throw new IllegalArgumentException("Invalid constant: " + obj);
        }
        return;
    }
    
    void checkLDCConstant(Object o) {
        if ((o instanceof Type)) {
            int sort = ((Type)o).getSort();
            if (sort != 10 && sort != 9 && sort != 11) {
                throw new IllegalArgumentException("Illegal LDC constant value");
            }
            if (sort != 11 && (this.version & 0xFFFF) < 49) {
                throw new IllegalArgumentException("ldc of a constant class requires at least version 1.5");
            }
            if (sort == 11 && (this.version & 0xFFFF) < 51) {
                throw new IllegalArgumentException("ldc of a method type requires at least version 1.7");
            }
        }
        else if ((o instanceof Handle)) {
            if ((this.version & 0xFFFF) < 51) {
                throw new IllegalArgumentException("ldc of a handle requires at least version 1.7");
            }
            int tag = ((Handle)o).getTag();
            if (tag < 1 || tag > 9) {
                throw new IllegalArgumentException("invalid handle tag " + tag);
            }
        }
        else {
            checkConstant(o);
        }
    }
    
    static void checkUnqualifiedName(int n, String str, String str2) {
        if ((n & 0xFFFF) < 49) {
            checkIdentifier(str, str2);
        }
        else {
            int i = 0;
            while (i < str.length()) {
                if (".;[/".indexOf((int)str.charAt(i)) != -1) {
                    throw new IllegalArgumentException("Invalid " + str2 + " (must be a valid unqualified name): " + str);
                }
                ++i;
            }
        }
    }
    
    static void checkIdentifier(String s, String s2) {
        checkIdentifier(s, 0, -1, s2);
    }
    
    static void checkIdentifier(String s, int index, int n, String str) {
        if (s == null || !((n != -1) ? (n > index) : (s.length() > index))) {
            throw new IllegalArgumentException("Invalid " + str + " (must not be null or empty)");
        }
        if (!Character.isJavaIdentifierStart(s.charAt(index))) {
            throw new IllegalArgumentException("Invalid " + str + " (must be a valid Java identifier): " + s);
        }
        n = ((n == -1) ? s.length() : n);
        int i = index + 1;
        while (i < n) {
            if (!Character.isJavaIdentifierPart(s.charAt(i))) {
                throw new IllegalArgumentException("Invalid " + str + " (must be a valid Java identifier): " + s);
            }
            ++i;
        }
        return;
    }
    
    static void checkMethodIdentifier(int n, String str, String s) {
        if (str == null || str.length() == 0) {
            throw new IllegalArgumentException("Invalid " + s + " (must not be null or empty)");
        }
        if ((n & 0xFFFF) >= 49) {
            int j = 0;
            while (j < str.length()) {
                if (".;[/<>".indexOf((int)str.charAt(j)) != -1) {
                    throw new IllegalArgumentException("Invalid " + s + " (must be a valid unqualified name): " + str);
                }
                ++j;
            }
            return;
        }
        if (!Character.isJavaIdentifierStart(str.charAt(0))) {
            throw new IllegalArgumentException("Invalid " + s + " (must be a '<init>', '<clinit>' or a valid Java identifier): " + str);
        }
        int i = 1;
        while (i < str.length()) {
            if (!Character.isJavaIdentifierPart(str.charAt(i))) {
                throw new IllegalArgumentException("Invalid " + s + " (must be '<init>' or '<clinit>' or a valid Java identifier): " + str);
            }
            ++i;
        }
        return;
    }
    
    static void checkInternalName(String s, String str) {
        if (s == null || s.length() == 0) {
            throw new IllegalArgumentException("Invalid " + str + " (must not be null or empty)");
        }
        if (s.charAt(0) == '[') {
            checkDesc(s, false);
        }
        else {
            checkInternalName(s, 0, -1, str);
        }
        return;
    }
    
    static void checkInternalName(String str, int n, int n2, String str2) {
        n2 = ((n2 == -1) ? str.length() : n2);
        try {
            int i;
            do {
                i = str.indexOf(47, n + 1);
                if (i == -1 || i > n2) {
                    i = n2;
                }
                checkIdentifier(str, n, i, null);
                n = i + 1;
            } while (i != n2);
        }
        catch (IllegalArgumentException ex) {
            throw new IllegalArgumentException("Invalid " + str2 + " (must be a fully qualified class name in internal form): " + str);
        }
    }
    
    static void checkDesc(String str, boolean b) {
        if (checkDesc(str, 0, b) != str.length()) {
            throw new IllegalArgumentException("Invalid descriptor: " + str);
        }
        return;
    }
    
    static int checkDesc(String str, int n, boolean b) {
        if (str == null || n >= str.length()) {
            throw new IllegalArgumentException("Invalid type descriptor (must not be null or empty)");
        }
        switch (str.charAt(n)) {
            case 'V': {
                if (b) {
                    return n + 1;
                }
                throw new IllegalArgumentException("Invalid descriptor: " + str);
            }
            case 'B':
            case 'C':
            case 'D':
            case 'F':
            case 'I':
            case 'J':
            case 'S':
            case 'Z': {
                return n + 1;
            }
            case '[': {
                int index;
                for (index = n + 1; index < str.length() && str.charAt(index) == '['; ++index) {}
                if (index < str.length()) {
                    return checkDesc(str, index, false);
                }
                throw new IllegalArgumentException("Invalid descriptor: " + str);
            }
            case 'L': {
                int index2 = str.indexOf(59, n);
                if (index2 == -1 || index2 - n < 2) {
                    throw new IllegalArgumentException("Invalid descriptor: " + str);
                }
                try {
                    checkInternalName(str, n + 1, index2, null);
                }
                catch (IllegalArgumentException ex) {
                    throw new IllegalArgumentException("Invalid descriptor: " + str);
                }
                return index2 + 1;
            }
            default: {
                throw new IllegalArgumentException("Invalid descriptor: " + str);
            }
        }
    }
    
    static void checkMethodDesc(String str) {
        if (str == null || str.length() == 0) {
            throw new IllegalArgumentException("Invalid method descriptor (must not be null or empty)");
        }
        if (str.charAt(0) != '(' || str.length() < 3) {
            throw new IllegalArgumentException("Invalid descriptor: " + str);
        }
        int checkDesc = 1;
        Label_0140: {
            if (str.charAt(checkDesc) != ')') {
                while (str.charAt(checkDesc) != 'V') {
                    checkDesc = checkDesc(str, checkDesc, false);
                    if (checkDesc >= str.length() || str.charAt(checkDesc) == ')') {
                        break Label_0140;
                    }
                    continue;
                }
                throw new IllegalArgumentException("Invalid descriptor: " + str);
            }
        }
        if (checkDesc(str, checkDesc + 1, true) != str.length()) {
            throw new IllegalArgumentException("Invalid descriptor: " + str);
        }
        return;
    }
    
    void checkLabel(Label label, boolean b, String s) {
        if (label == null) {
            throw new IllegalArgumentException("Invalid " + s + " (must not be null)");
        }
        if (b && this.labels.get(label) == null) {
            throw new IllegalArgumentException("Invalid " + s + " (must be visited first)");
        }
        return;
    }
    
    private static void checkNonDebugLabel(Label obj) {
        Field labelStatusField = getLabelStatusField();
        int n;
        try {
            n = ((labelStatusField == null) ? 0 : ((int)labelStatusField.get(obj)));
        }
        catch (IllegalAccessException ex) {
            throw new Error("Internal error");
        }
        if ((n & 0x1) != 0x0) {
            throw new IllegalArgumentException("Labels used for debug info cannot be reused for control flow");
        }
        return;
    }
    
    private static Field getLabelStatusField() {
        if (CheckMethodAdapter.labelStatusField == null) {
            CheckMethodAdapter.labelStatusField = getLabelField("a");
            if (CheckMethodAdapter.labelStatusField == null) {
                CheckMethodAdapter.labelStatusField = getLabelField("status");
            }
        }
        return CheckMethodAdapter.labelStatusField;
    }
    
    private static Field getLabelField(String name) {
        try {
            Field declaredField = Label.class.getDeclaredField(name);
            declaredField.setAccessible(true);
            return declaredField;
        }
        catch (NoSuchFieldException ex) {
            return null;
        }
    }
    
    static {
        String s = "BBBBBBBBBBBBBBBBCCIAADDDDDAAAAAAAAAAAAAAAAAAAABBBBBBBBDDDDDAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBJBBBBBBBBBBBBBBBBBBBBHHHHHHHHHHHHHHHHDKLBBBBBBFFFFGGGGAECEBBEEBBAMHHAA";
        TYPE = new int[s.length()];
        for (int i = 0; i < CheckMethodAdapter.TYPE.length; ++i) {
            CheckMethodAdapter.TYPE[i] = s.charAt(i) - 65 - 1;
        }
    }
}
